@@ -26,12 +26,12 @@
 #include "format.h"
 #include "c-ctype.h"
 #include "gcd.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "xerror.h"
 #include "format-invalid.h"
 #include "minmax.h"
 #include "error.h"
-#include "progname.h"
+#include "error-progname.h"
 #include "gettext.h"
 
 #define _(str) gettext (str)
@@ -417,7 +417,7 @@ normalize_outermost_list (struct format_arg_list *list)
   n = list->initial.count;
   for (i = j = 0; i < n; i++)
     if (j > 0
-        && equal_element (&list->initial.element[i],
+	&& equal_element (&list->initial.element[i],
 			  &list->initial.element[j-1]))
       {
 	list->initial.element[j-1].repcount +=
@@ -435,7 +435,7 @@ normalize_outermost_list (struct format_arg_list *list)
   n = list->repeated.count;
   for (i = j = 0; i < n; i++)
     if (j > 0
-        && equal_element (&list->repeated.element[i],
+	&& equal_element (&list->repeated.element[i],
 			  &list->repeated.element[j-1]))
       {
 	list->repeated.element[j-1].repcount +=
@@ -2386,7 +2386,10 @@ check_params (struct format_arg_list **listp,
       case PT_CHARACTER: case PT_INTEGER: case PT_ARGCOUNT:
 	/* too many params for directive */
 	*invalid_reason =
-	  xasprintf (_("In the directive number %u, too many parameters are given; expected at most %u parameters."), directives, orig_t_count);
+	  xasprintf (ngettext ("In the directive number %u, too many parameters are given; expected at most %u parameter.",
+			       "In the directive number %u, too many parameters are given; expected at most %u parameters.",
+			       orig_t_count),
+		     directives, orig_t_count);
 	return false;
       case PT_V:
 	/* Force argument to be NIL.  */
@@ -2469,7 +2472,7 @@ parse_upto (const char **formatp,
 	unsigned int paramcount = 0;
 	struct param *params = NULL;
 
-        /* Count number of directives.  */
+	/* Count number of directives.  */
 	spec->directives++;
 
 	/* Parse parameters.  */
