@@ -29,12 +29,14 @@
 
 #include "fstrcmp.h"
 #include "hash.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 
 
 const char *const format_language[NFORMATS] =
 {
   /* format_c */		"c",
+  /* format_objc */		"objc",
+  /* format_sh */		"sh",
   /* format_python */		"python",
   /* format_lisp */		"lisp",
   /* format_elisp */		"elisp",
@@ -45,12 +47,18 @@ const char *const format_language[NFORMATS] =
   /* format_pascal */		"object-pascal",
   /* format_ycp */		"ycp",
   /* format_tcl */		"tcl",
-  /* format_php */		"php"
+  /* format_perl */		"perl",
+  /* format_perl_brace */	"perl-brace",
+  /* format_php */		"php",
+  /* format_gcc_internal */	"gcc-internal",
+  /* format_qt */		"qt"
 };
 
 const char *const format_language_pretty[NFORMATS] =
 {
   /* format_c */		"C",
+  /* format_objc */		"Objective C",
+  /* format_sh */		"Shell",
   /* format_python */		"Python",
   /* format_lisp */		"Lisp",
   /* format_elisp */		"Emacs Lisp",
@@ -61,14 +69,20 @@ const char *const format_language_pretty[NFORMATS] =
   /* format_pascal */		"Object Pascal",
   /* format_ycp */		"YCP",
   /* format_tcl */		"Tcl",
-  /* format_php */		"PHP"
+  /* format_perl */		"Perl",
+  /* format_perl_brace */	"Perl brace",
+  /* format_php */		"PHP",
+  /* format_gcc_internal */	"GCC internal",
+  /* format_qt */		"Qt"
 };
 
 
 bool
 possible_format_p (enum is_format is_format)
 {
-  return is_format == possible || is_format == yes;
+  return is_format == possible
+	 || is_format == yes_according_to_context
+	 || is_format == yes;
 }
 
 
@@ -316,17 +330,48 @@ message_list_remove_if_not (message_list_ty *mlp,
 }
 
 
+bool
+message_list_msgids_changed (message_list_ty *mlp)
+{
+  if (mlp->use_hashtable)
+    {
+      unsigned long int size = mlp->htable.size;
+      size_t j;
+
+      delete_hash (&mlp->htable);
+      init_hash (&mlp->htable, size);
+
+      for (j = 0; j < mlp->nitems; j++)
+	{
+	  message_ty *mp = mlp->item[j];
+
+	  if (insert_entry (&mlp->htable, mp->msgid, strlen (mp->msgid) + 1,
+			    mp))
+	    /* A message list has duplicates, although it was allocated with
+	       the assertion that it wouldn't have duplicates, and before the
+	       msgids changed it indeed didn't have duplicates.  */
+	    {
+	      delete_hash (&mlp->htable);
+	      mlp->use_hashtable = false;
+	      return true;
+	    }
+	}
+    }
+  return false;
+}
+
+
 message_ty *
 message_list_search (message_list_ty *mlp, const char *msgid)
 {
   if (mlp->use_hashtable)
     {
-      message_ty *mp;
+      void *htable_value;
 
-      if (find_entry (&mlp->htable, msgid, strlen (msgid) + 1, (void **) &mp))
+      if (find_entry (&mlp->htable, msgid, strlen (msgid) + 1, &htable_value))
 	return NULL;
       else
-	return mp;
+	return (message_ty *) htable_value;
     }
   else
     {
@@ -604,7 +649,7 @@ msgdomain_list_search (msgdomain_list_ty *mdlp, const char *msgid)
       mdp = mdlp->item[j];
       mp = message_list_search (mdp->messages, msgid);
       if (mp)
-        return mp;
+	return mp;
     }
   return NULL;
 }
