@@ -37,6 +37,8 @@
 
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
+#elif defined _MSC_VER || defined __MINGW32__
+# include <io.h>
 #endif
 
 /* Get fd_set (on AIX) or select() declaration (on EMX).  */
@@ -44,8 +46,10 @@
 # include <sys/select.h>
 #endif
 
+#include "closeout.h"
 #include "dir-list.h"
 #include "error.h"
+#include "error-progname.h"
 #include "progname.h"
 #include "relocatable.h"
 #include "basename.h"
@@ -53,7 +57,7 @@
 #include "read-po.h"
 #include "write-po.h"
 #include "msgl-charset.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "exit.h"
 #include "findprog.h"
 #include "pipe.h"
@@ -112,6 +116,8 @@ static const struct option long_options[] =
   { "sort-by-file", no_argument, NULL, 'F' },
   { "sort-output", no_argument, NULL, 's' },
   { "strict", no_argument, NULL, 'S' },
+  { "stringtable-input", no_argument, NULL, CHAR_MAX + 4 },
+  { "stringtable-output", no_argument, NULL, CHAR_MAX + 5 },
   { "version", no_argument, NULL, 'V' },
   { "width", required_argument, NULL, 'w', },
   { NULL, 0, NULL, 0 }
@@ -153,6 +159,9 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, relocate (LOCALEDIR));
   textdomain (PACKAGE);
 
+  /* Ensure that write errors on stdout are detected.  */
+  atexit (close_stdout);
+
   /* Set default values for variables.  */
   do_help = false;
   do_version = false;
@@ -240,6 +249,14 @@ main (int argc, char **argv)
 	message_page_width_ignore ();
 	break;
 
+      case CHAR_MAX + 4: /* --stringtable-input */
+	input_syntax = syntax_stringtable;
+	break;
+
+      case CHAR_MAX + 5: /* --stringtable-output */
+	message_print_syntax_stringtable ();
+	break;
+
       default:
 	usage (EXIT_FAILURE);
 	break;
@@ -398,6 +415,8 @@ Useful FILTER-OPTIONs when the FILTER is 'sed':\n"));
 Input file syntax:\n"));
       printf (_("\
   -P, --properties-input      input file is in Java .properties syntax\n"));
+      printf (_("\
+      --stringtable-input     input file is in NeXTstep/GNUstep .strings syntax\n"));
       printf ("\n");
       printf (_("\
 Output details:\n"));
@@ -420,6 +439,8 @@ Output details:\n"));
       printf (_("\
   -p, --properties-output     write out a Java .properties file\n"));
       printf (_("\
+      --stringtable-output    write out a NeXTstep/GNUstep .strings file\n"));
+      printf (_("\
   -w, --width=NUMBER          set output page width\n"));
       printf (_("\
       --no-wrap               do not break long message lines, longer than\n\
@@ -546,7 +567,8 @@ process_string (const char *str, size_t len, char **resultp, size_t *lengthp)
   int exitstatus;
 
   /* Open a bidirectional pipe to a subprocess.  */
-  child = create_pipe_bidi (sub_name, sub_path, sub_argv, false, true, fd);
+  child = create_pipe_bidi (sub_name, sub_path, sub_argv, false, true, true,
+			    fd);
 
   /* Enable non-blocking I/O.  This permits the read() and write() calls
      to return -1/EAGAIN without blocking; this is important for polling
@@ -637,7 +659,7 @@ process_string (const char *str, size_t len, char **resultp, size_t *lengthp)
       if (length == allocated)
 	{
 	  allocated = allocated + (allocated >> 1);
-	  result = xrealloc (result, allocated);
+	  result = (char *) xrealloc (result, allocated);
 	}
       {
 	ssize_t nread = read (fd[0], result + length, allocated - length);
@@ -657,7 +679,7 @@ process_string (const char *str, size_t len, char **resultp, size_t *lengthp)
   close (fd[0]);
 
   /* Remove zombie process from process list.  */
-  exitstatus = wait_subprocess (child, sub_name, true);
+  exitstatus = wait_subprocess (child, sub_name, false, true, true);
   if (exitstatus != 0)
     error (EXIT_FAILURE, 0, _("%s subprocess terminated with exit code %d"),
 	   sub_name, exitstatus);
@@ -717,7 +739,7 @@ process_message (message_ty *mp)
       size_t length;
 
       process_string (p, strlen (p), &result, &length);
-      result = xrealloc (result, length + 1);
+      result = (char *) xrealloc (result, length + 1);
       result[length] = '\0';
       substrings[k] = result;
       total_len += length + 1;
