@@ -30,14 +30,16 @@
 #include <string.h>
 #include <locale.h>
 
+#include "closeout.h"
 #include "dir-list.h"
 #include "error.h"
+#include "error-progname.h"
 #include "progname.h"
 #include "relocatable.h"
 #include "basename.h"
 #include "xerror.h"
 #include "format.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "plural-exp.h"
 #include "plural-table.h"
 #include "strstr.h"
@@ -47,6 +49,7 @@
 #include "write-mo.h"
 #include "write-java.h"
 #include "write-tcl.h"
+#include "write-qt.h"
 
 #include "gettext.h"
 #include "message.h"
@@ -93,6 +96,9 @@ static bool tcl_mode;
 static const char *tcl_locale_name;
 static const char *tcl_base_directory;
 
+/* Qt mode output file specification.  */
+static bool qt_mode;
+
 /* We may have more than one input file.  Domains with same names in
    different files have to merged.  So we need a list of tables for
    each output file.  */
@@ -162,9 +168,11 @@ static const struct option long_options[] =
   { "no-hash", no_argument, NULL, CHAR_MAX + 6 },
   { "output-file", required_argument, NULL, 'o' },
   { "properties-input", no_argument, NULL, 'P' },
+  { "qt", no_argument, NULL, CHAR_MAX + 9 },
   { "resource", required_argument, NULL, 'r' },
   { "statistics", no_argument, &do_statistics, 1 },
   { "strict", no_argument, NULL, 'S' },
+  { "stringtable-input", no_argument, NULL, CHAR_MAX + 8 },
   { "tcl", no_argument, NULL, CHAR_MAX + 7 },
   { "use-fuzzy", no_argument, NULL, 'f' },
   { "verbose", no_argument, NULL, 'v' },
@@ -214,6 +222,9 @@ main (int argc, char *argv[])
   bindtextdomain (PACKAGE, relocate (LOCALEDIR));
   textdomain (PACKAGE);
 
+  /* Ensure that write errors on stdout are detected.  */
+  atexit (close_stdout);
+
   while ((opt = getopt_long (argc, argv, "a:cCd:D:fhjl:o:Pr:vV", long_options,
 			     NULL))
 	 != EOF)
@@ -276,7 +287,7 @@ main (int argc, char *argv[])
       case 'V':
 	do_version = true;
 	break;
-      case CHAR_MAX + 1:
+      case CHAR_MAX + 1: /* --check-accelerators */
 	check_accelerators = true;
 	if (optarg != NULL)
 	  {
@@ -289,25 +300,31 @@ main (int argc, char *argv[])
 		     "--check-accelerators");
 	  }
 	break;
-      case CHAR_MAX + 2:
+      case CHAR_MAX + 2: /* --check-domain */
 	check_domain = true;
 	break;
-      case CHAR_MAX + 3:
+      case CHAR_MAX + 3: /* --check-format */
 	check_format_strings = true;
 	break;
-      case CHAR_MAX + 4:
+      case CHAR_MAX + 4: /* --check-header */
 	check_header = true;
 	break;
-      case CHAR_MAX + 5:
+      case CHAR_MAX + 5: /* --java2 */
 	java_mode = true;
 	assume_java2 = true;
 	break;
-      case CHAR_MAX + 6:
+      case CHAR_MAX + 6: /* --no-hash */
 	no_hash_table = true;
 	break;
-      case CHAR_MAX + 7:
+      case CHAR_MAX + 7: /* --tcl */
 	tcl_mode = true;
 	break;
+      case CHAR_MAX + 8: /* --stringtable-input */
+	input_syntax = syntax_stringtable;
+	break;
+      case CHAR_MAX + 9: /* --qt */
+	qt_mode = true;
+	break;
       default:
 	usage (EXIT_FAILURE);
 	break;
@@ -342,6 +359,12 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
   if (java_mode && tcl_mode)
     error (EXIT_FAILURE, 0, _("%s and %s are mutually exclusive"),
 	   "--java", "--tcl");
+  if (java_mode && qt_mode)
+    error (EXIT_FAILURE, 0, _("%s and %s are mutually exclusive"),
+	   "--java", "--qt");
+  if (tcl_mode && qt_mode)
+    error (EXIT_FAILURE, 0, _("%s and %s are mutually exclusive"),
+	   "--tcl", "--qt");
   if (java_mode)
     {
       if (output_file_name != NULL)
@@ -406,8 +429,8 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
   if (output_file_name != NULL)
     current_domain =
       new_domain (output_file_name,
-		  strict_uniforum ? add_mo_suffix (output_file_name)
-				  : output_file_name);
+		  !qt_mode && strict_uniforum ? add_mo_suffix (output_file_name)
+					      : output_file_name);
 
   /* Process all given .po files.  */
   while (argc > optind)
@@ -423,8 +446,12 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
       ++optind;
     }
 
-  /* We know a priori that properties_parse() converts strings to UTF-8.  */
-  canon_encoding = (input_syntax == syntax_properties ? po_charset_utf8 : NULL);
+  /* We know a priori that properties_parse() and stringtable_parse() convert
+     strings to UTF-8.  */
+  canon_encoding =
+    (input_syntax == syntax_properties || input_syntax == syntax_stringtable
+     ? po_charset_utf8
+     : NULL);
 
   /* Remove obsolete messages.  They were only needed for duplicate
      checking.  */
@@ -452,6 +479,12 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 				   tcl_locale_name, tcl_base_directory))
 	    exit_status = EXIT_FAILURE;
 	}
+      else if (qt_mode)
+	{
+	  if (msgdomain_write_qt (domain->mlp, canon_encoding,
+				  domain->domain_name, domain->file_name))
+	    exit_status = EXIT_FAILURE;
+	}
       else
 	{
 	  if (msgdomain_write_mo (domain->mlp, domain->domain_name,
@@ -528,6 +561,8 @@ Operation mode:\n"));
       --java2                 like --java, and assume Java2 (JDK 1.2 or higher)\n"));
       printf (_("\
       --tcl                   Tcl mode: generate a tcl/msgcat .msg file\n"));
+      printf (_("\
+      --qt                    Qt mode: generate a Qt .qm file\n"));
       printf ("\n");
       printf (_("\
 Output file location:\n"));
@@ -566,6 +601,9 @@ specified directory.\n"));
 Input file syntax:\n"));
       printf (_("\
   -P, --properties-input      input files are in Java .properties syntax\n"));
+      printf (_("\
+      --stringtable-input     input files are in NeXTstep/GNUstep .strings\n\
+                              syntax\n"));
       printf ("\n");
       printf (_("\
 Input file interpretation:\n"));
@@ -779,16 +817,19 @@ check_plural_eval (struct expression *plural_expr,
 #if USE_SIGINFO
 # ifdef FPE_INTDIV
 	case FPE_INTDIV:
+	  /* xgettext: c-format */
 	  msg = _("plural expression can produce division by zero");
 	  break;
 # endif
 # ifdef FPE_INTOVF
 	case FPE_INTOVF:
+	  /* xgettext: c-format */
 	  msg = _("plural expression can produce integer overflow");
 	  break;
 # endif
 	default:
 #endif
+	  /* xgettext: c-format */
 	  msg = _("plural expression can produce arithmetic exceptions, possibly division by zero");
 	}
 
@@ -1372,7 +1413,7 @@ msgfmt_set_domain (default_po_reader_ty *this, char *name)
 {
   /* If no output file was given, we change it with each `domain'
      directive.  */
-  if (!java_mode && !tcl_mode && output_file_name == NULL)
+  if (!java_mode && !tcl_mode && !qt_mode && output_file_name == NULL)
     {
       size_t correct;
 
@@ -1409,7 +1450,7 @@ domain name \"%s\" not suitable as file name: will use prefix"), name);
 }
 
 
-void
+static void
 msgfmt_add_message (default_po_reader_ty *this,
 		    char *msgid,
 		    lex_pos_ty *msgid_pos,
