@@ -69,7 +69,9 @@
 # define HAVE_DIR 0
 #endif
 
+#include "closeout.h"
 #include "error.h"
+#include "error-progname.h"
 #include "progname.h"
 #include "relocatable.h"
 #include "basename.h"
@@ -83,7 +85,7 @@
 #include "localcharset.h"
 #include "po-time.h"
 #include "plural-table.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "exit.h"
 #include "pathname.h"
 #include "xerror.h"
@@ -131,6 +133,8 @@ static const struct option long_options[] =
   { "output-file", required_argument, NULL, 'o' },
   { "properties-input", no_argument, NULL, 'P' },
   { "properties-output", no_argument, NULL, 'p' },
+  { "stringtable-input", no_argument, NULL, CHAR_MAX + 3 },
+  { "stringtable-output", no_argument, NULL, CHAR_MAX + 4 },
   { "version", no_argument, NULL, 'V' },
   { "width", required_argument, NULL, 'w' },
   { NULL, 0, NULL, 0 }
@@ -172,6 +176,9 @@ main (int argc, char **argv)
   bindtextdomain (PACKAGE, relocate (LOCALEDIR));
   textdomain (PACKAGE);
 
+  /* Ensure that write errors on stdout are detected.  */
+  atexit (close_stdout);
+
   /* Set default values for variables.  */
   do_help = false;
   do_version = false;
@@ -237,6 +244,14 @@ main (int argc, char **argv)
 	message_page_width_ignore ();
 	break;
 
+      case CHAR_MAX + 3: /* --stringtable-input */
+	input_syntax = syntax_stringtable;
+	break;
+
+      case CHAR_MAX + 4: /* --stringtable-output */
+	message_print_syntax_stringtable ();
+	break;
+
       default:
 	usage (EXIT_FAILURE);
 	break;
@@ -325,7 +340,8 @@ the output .po file through the --output-file option.\n"),
   /* Write the modified message list out.  */
   msgdomain_list_print (result, output_file, true, false);
 
-  fprintf (stderr, "\n");
+  if (!no_translator)
+    fprintf (stderr, "\n");
   fprintf (stderr, _("Created %s.\n"), output_file);
 
   exit (EXIT_SUCCESS);
@@ -374,6 +390,8 @@ locale setting.  If it is -, the results are written to standard output.\n"));
 Input file syntax:\n"));
       printf (_("\
   -P, --properties-input      input file is in Java .properties syntax\n"));
+      printf (_("\
+      --stringtable-input     input file is in NeXTstep/GNUstep .strings syntax\n"));
       printf ("\n");
       printf (_("\
 Output details:\n"));
@@ -384,6 +402,8 @@ Output details:\n"));
       printf (_("\
   -p, --properties-output     write out a Java .properties file\n"));
       printf (_("\
+      --stringtable-output    write out a NeXTstep/GNUstep .strings file\n"));
+      printf (_("\
   -w, --width=NUMBER          set output page width\n"));
       printf (_("\
       --no-wrap               do not break long message lines, longer than\n\
@@ -476,6 +496,7 @@ catalogname_for_locale (const char *locale)
 		/* Language	Main territory */
     "af_ZA",	/* Afrikaans	South Africa */
     "am_ET",	/* Amharic	Ethiopia */
+    "an_ES",	/* Aragonese	Spain */
     "as_IN",	/* Assamese	India */
     "az_AZ",	/* Azerbaijani	Azerbaijan */
     "be_BY",	/* Belarusian	Belarus */
@@ -508,9 +529,11 @@ catalogname_for_locale (const char *locale)
     "he_IL",	/* Hebrew	Israel */
     "hi_IN",	/* Hindi	India */
     "hr_HR",	/* Croatian	Croatia */
+    "ht_HT",	/* Haitian	Haiti */
     "hu_HU",	/* Hungarian	Hungary */
     "hy_AM",	/* Armenian	Armenia */
     "id_ID",	/* Indonesian	Indonesia */
+    "ii_CN",	/* Sichuan Yi	China */
     "is_IS",	/* Icelandic	Iceland */
     "it_IT",	/* Italian	Italy */
     "ja_JP",	/* Japanese	Japan */
@@ -522,6 +545,7 @@ catalogname_for_locale (const char *locale)
     "kn_IN",	/* Kannada	India */
     "ko_KR",	/* Korean	Korea (South) */
     "kok_IN",	/* Konkani	India */
+    "li_BE",	/* Limburgish	Belgium */
     "lo_LA",	/* Laotian	Laos */
     "lt_LT",	/* Lithuanian	Lithuania */
     "lv_LV",	/* Latvian	Latvia */
@@ -698,6 +722,7 @@ englishname_of_language ()
     { "ae", "Avestan" },
     { "af", "Afrikaans" },
     { "am", "Amharic" },
+    { "an", "Aragonese" },
     { "ar", "Arabic" },
     { "as", "Assamese" },
     { "ay", "Aymara" },
@@ -745,12 +770,14 @@ englishname_of_language ()
     { "hi", "Hindi" },
     { "ho", "Hiri Motu" },
     { "hr", "Croatian" },
+    { "ht", "Haitian" },
     { "hu", "Hungarian" },
     { "hy", "Armenian" },
     { "hz", "Herero" },
     { "ia", "Interlingua" },
     { "id", "Indonesian" },
     { "ie", "Interlingue" },
+    { "ii", "Sichuan Yi" },
     { "ik", "Inupiak" },
     { "is", "Icelandic" },
     { "it", "Italian" },
@@ -773,6 +800,7 @@ englishname_of_language ()
     { "kok", "Konkani" },
     { "la", "Latin" },
     { "lb", "Letzeburgesch" },
+    { "li", "Limburgish" },
     { "ln", "Lingala" },
     { "lo", "Laotian" },
     { "lt", "Lithuanian" },
@@ -897,29 +925,44 @@ project_id ()
   argv[0] = "/bin/sh";
   argv[1] = prog;
   argv[2] = NULL;
-  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, fd);
+  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, false,
+			  fd);
+  if (child == -1)
+    goto failed;
 
   /* Retrieve its result.  */
   fp = fdopen (fd[0], "r");
   if (fp == NULL)
-    error (EXIT_FAILURE, errno, _("fdopen() failed"));
+    {
+      error (0, errno, _("fdopen() failed"));
+      goto failed;
+    }
 
   line = NULL; linesize = 0;
   linelen = getline (&line, &linesize, fp);
   if (linelen == (size_t)(-1))
-    error (EXIT_FAILURE, 0, _("%s subprocess I/O error"), prog);
+    {
+      error (0, 0, _("%s subprocess I/O error"), prog);
+      goto failed;
+    }
   if (linelen > 0 && line[linelen - 1] == '\n')
     line[linelen - 1] = '\0';
 
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, true);
+  exitstatus = wait_subprocess (child, prog, false, true, false);
   if (exitstatus != 0)
-    error (EXIT_FAILURE, 0, _("%s subprocess failed with exit code %d"),
-	   prog, exitstatus);
+    {
+      error (0, 0, _("%s subprocess failed with exit code %d"),
+	     prog, exitstatus);
+      goto failed;
+    }
 
   return line;
+
+failed:
+  return "PACKAGE";
 }
 
 
@@ -949,29 +992,44 @@ project_id_version ()
   argv[1] = prog;
   argv[2] = "yes";
   argv[3] = NULL;
-  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, fd);
+  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, false,
+			  fd);
+  if (child == -1)
+    goto failed;
 
   /* Retrieve its result.  */
   fp = fdopen (fd[0], "r");
   if (fp == NULL)
-    error (EXIT_FAILURE, errno, _("fdopen() failed"));
+    {
+      error (0, errno, _("fdopen() failed"));
+      goto failed;
+    }
 
   line = NULL; linesize = 0;
   linelen = getline (&line, &linesize, fp);
   if (linelen == (size_t)(-1))
-    error (EXIT_FAILURE, 0, _("%s subprocess I/O error"), prog);
+    {
+      error (0, 0, _("%s subprocess I/O error"), prog);
+      goto failed;
+    }
   if (linelen > 0 && line[linelen - 1] == '\n')
     line[linelen - 1] = '\0';
 
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, true);
+  exitstatus = wait_subprocess (child, prog, false, true, false);
   if (exitstatus != 0)
-    error (EXIT_FAILURE, 0, _("%s subprocess failed with exit code %d"),
-	   prog, exitstatus);
+    {
+      error (0, 0, _("%s subprocess failed with exit code %d"),
+	     prog, exitstatus);
+      goto failed;
+    }
 
   return line;
+
+failed:
+  return "PACKAGE VERSION";
 }
 
 
@@ -1092,29 +1150,44 @@ The new message catalog should contain your email address, so that users can\n\
 give you feedback about the translations, and so that maintainers can contact\n\
 you in case of unexpected technical problems.\n");
   argv[3] = NULL;
-  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, fd);
+  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, false,
+			  fd);
+  if (child == -1)
+    goto failed;
 
   /* Retrieve his answer.  */
   fp = fdopen (fd[0], "r");
   if (fp == NULL)
-    error (EXIT_FAILURE, errno, _("fdopen() failed"));
+    {
+      error (0, errno, _("fdopen() failed"));
+      goto failed;
+    }
 
   line = NULL; linesize = 0;
   linelen = getline (&line, &linesize, fp);
   if (linelen == (size_t)(-1))
-    error (EXIT_FAILURE, 0, _("%s subprocess I/O error"), prog);
+    {
+      error (0, 0, _("%s subprocess I/O error"), prog);
+      goto failed;
+    }
   if (linelen > 0 && line[linelen - 1] == '\n')
     line[linelen - 1] = '\0';
 
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, true);
+  exitstatus = wait_subprocess (child, prog, false, true, false);
   if (exitstatus != 0)
-    error (EXIT_FAILURE, 0, _("%s subprocess failed with exit code %d"),
-	   prog, exitstatus);
+    {
+      error (0, 0, _("%s subprocess failed with exit code %d"),
+	     prog, exitstatus);
+      goto failed;
+    }
 
   return line;
+
+failed:
+  return "EMAIL@ADDRESS";
 }
 
 
@@ -1159,12 +1232,18 @@ language_team_address ()
   argv[4] = (char *) catalogname;
   argv[5] = (char *) language;
   argv[6] = NULL;
-  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, fd);
+  child = create_pipe_in (prog, "/bin/sh", argv, DEV_NULL, false, true, false,
+			  fd);
+  if (child == -1)
+    goto failed;
 
   /* Retrieve its result.  */
   fp = fdopen (fd[0], "r");
   if (fp == NULL)
-    error (EXIT_FAILURE, errno, _("fdopen() failed"));
+    {
+      error (0, errno, _("fdopen() failed"));
+      goto failed;
+    }
 
   line = NULL; linesize = 0;
   linelen = getline (&line, &linesize, fp);
@@ -1176,12 +1255,18 @@ language_team_address ()
   fclose (fp);
 
   /* Remove zombie process from process list, and retrieve exit status.  */
-  exitstatus = wait_subprocess (child, prog, true);
+  exitstatus = wait_subprocess (child, prog, false, true, false);
   if (exitstatus != 0)
-    error (EXIT_FAILURE, 0, _("%s subprocess failed with exit code %d"),
-	   prog, exitstatus);
+    {
+      error (0, 0, _("%s subprocess failed with exit code %d"),
+	     prog, exitstatus);
+      goto failed;
+    }
 
   return line;
+
+failed:
+  return "";
 }
 
 
