@@ -30,7 +30,6 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <signal.h>
 
 #include <sys/stat.h>
 #if STAT_MACROS_BROKEN
@@ -62,6 +61,12 @@
 # include <unistd.h>
 #endif
 
+#ifdef __MINGW32__
+/* mingw's mkdir() function has 1 argument, but we pass 2 arguments.
+   Therefore we have to disable the argument count checking.  */
+# define mkdir ((int (*)()) mkdir)
+#endif
+
 #include "c-ctype.h"
 #include "error.h"
 #include "javacomp.h"
@@ -72,8 +77,10 @@
 #include "pathmax.h"
 #include "plural-exp.h"
 #include "po-charset.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "pathname.h"
+#include "fatal-signal.h"
+#include "fwriteerror.h"
 #include "tmpdir.h"
 #include "utf8-ucs4.h"
 #include "gettext.h"
@@ -893,26 +900,9 @@ static struct
   const char *file_name;
 } cleanup_list;
 
-static void
-uninstall_handlers ()
-{
-#ifdef SIGHUP
-  signal (SIGHUP, SIG_DFL);
-#endif
-#ifdef SIGINT
-  signal (SIGINT, SIG_DFL);
-#endif
-#ifdef SIGPIPE
-  signal (SIGPIPE, SIG_DFL);
-#endif
-#ifdef SIGTERM
-  signal (SIGTERM, SIG_DFL);
-#endif
-}
-
 /* The signal handler.  It gets called asynchronously.  */
 static void
-cleanup (int sig)
+cleanup ()
 {
   unsigned int i;
 
@@ -939,83 +929,8 @@ cleanup (int sig)
     if (filename != NULL)
       rmdir (filename);
   }
-
-  /* Now execute the signal's default action.  */
-  uninstall_handlers ();
-#if HAVE_RAISE
-  raise (sig);
-#else
-  kill (getpid (), sig);
-#endif
 }
 
-static void
-install_handlers ()
-{
-#ifdef SIGHUP
-  signal (SIGHUP, &cleanup);
-#endif
-#ifdef SIGINT
-  signal (SIGINT, &cleanup);
-#endif
-#ifdef SIGPIPE
-  signal (SIGPIPE, &cleanup);
-#endif
-#ifdef SIGTERM
-  signal (SIGTERM, &cleanup);
-#endif
-}
-
-#if HAVE_POSIX_SIGNALBLOCKING
-
-static sigset_t signal_set;
-
-static void
-init_signal_set ()
-{
-  static bool signal_set_initialized = false;
-  if (!signal_set_initialized)
-    {
-      sigemptyset (&signal_set);
-
-#ifdef SIGHUP
-      sigaddset (&signal_set, SIGHUP);
-#endif
-#ifdef SIGINT
-      sigaddset (&signal_set, SIGINT);
-#endif
-#ifdef SIGPIPE
-      sigaddset (&signal_set, SIGPIPE);
-#endif
-#ifdef SIGTERM
-      sigaddset (&signal_set, SIGTERM);
-#endif
-
-      signal_set_initialized = true;
-    }
-}
-
-static void
-block ()
-{
-  init_signal_set ();
-  sigprocmask (SIG_BLOCK, &signal_set, NULL);
-}
-
-static void
-unblock ()
-{
-  init_signal_set ();
-  sigprocmask (SIG_UNBLOCK, &signal_set, NULL);
-}
-
-#else
-
-#define block() /* nothing */
-#define unblock() /* nothing */
-
-#endif
-
 
 int
 msgdomain_write_java (message_list_ty *mlp, const char *canon_encoding,
@@ -1045,7 +960,14 @@ msgdomain_write_java (message_list_ty *mlp, const char *canon_encoding,
   cleanup_list.tmpdir = NULL;
   cleanup_list.subdir_count = 0;
   cleanup_list.file_name = NULL;
-  install_handlers ();
+  {
+    static bool cleanup_already_registered = false;
+    if (!cleanup_already_registered)
+      {
+	at_fatal_signal (&cleanup);
+	cleanup_already_registered = true;
+      }
+  }
 
   /* Create a temporary directory where we can put the Java file.  */
   template = (char *) alloca (PATH_MAX);
@@ -1055,10 +977,10 @@ msgdomain_write_java (message_list_ty *mlp, const char *canon_encoding,
 	     _("cannot find a temporary directory, try setting $TMPDIR"));
       goto quit1;
     }
-  block ();
+  block_fatal_signals ();
   tmpdir = mkdtemp (template);
   cleanup_list.tmpdir = tmpdir;
-  unblock ();
+  unblock_fatal_signals ();
   if (tmpdir == NULL)
     {
       error (0, errno,
@@ -1151,7 +1073,7 @@ msgdomain_write_java (message_list_ty *mlp, const char *canon_encoding,
 
   write_java_code (java_file, class_name, mlp, assume_java2);
 
-  if (fflush (java_file) || ferror (java_file))
+  if (fwriteerror (java_file))
     {
       error (0, errno, _("error while writing \"%s\" file"), java_file_name);
       fclose (java_file);
@@ -1196,6 +1118,6 @@ compilation of Java class failed, please try --verbose or set $JAVAC"));
   rmdir (tmpdir);
  quit1:
   cleanup_list.tmpdir = NULL;
-  uninstall_handlers ();
+  /* Here we could unregister the cleanup() handler.  */
   return retval;
 }
