@@ -42,8 +42,9 @@
 #include "hash.h"
 #include "message.h"
 #include "format.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "binary-io.h"
+#include "fwriteerror.h"
 #include "exit.h"
 #include "gettext.h"
 
@@ -180,7 +181,8 @@ write_table (FILE *output_file, message_list_ty *mlp)
 
       /* Test if mp contains system dependent strings and thus
 	 requires the use of the .mo file minor revision 1.  */
-      if (possible_format_p (mp->is_format[format_c]))
+      if (possible_format_p (mp->is_format[format_c])
+	  || possible_format_p (mp->is_format[format_objc]))
 	{
 	  /* Check whether msgid or msgstr contain ISO C 99 <inttypes.h>
 	     format string directives.  No need to check msgid_plural, because
@@ -701,7 +703,7 @@ msgdomain_write_mo (message_list_ty *mlp,
 	  write_table (output_file, mlp);
 
 	  /* Make sure nothing went wrong.  */
-	  if (fflush (output_file) || ferror (output_file))
+	  if (fwriteerror (output_file))
 	    error (EXIT_FAILURE, errno, _("error while writing \"%s\" file"),
 		   file_name);
 
