@@ -1,5 +1,5 @@
 /* xgettext C/C++/ObjectiveC backend.
-   Copyright (C) 1995-1998, 2000-2002 Free Software Foundation, Inc.
+   Copyright (C) 1995-1998, 2000-2003 Free Software Foundation, Inc.
 
    This file was written by Peter Miller <millerp@canb.auug.org.au>
 
@@ -28,11 +28,11 @@
 #include <string.h>
 
 #include "message.h"
-#include "x-c.h"
 #include "xgettext.h"
+#include "x-c.h"
 #include "error.h"
-#include "progname.h"
-#include "xmalloc.h"
+#include "error-progname.h"
+#include "xalloc.h"
 #include "exit.h"
 #include "hash.h"
 #include "gettext.h"
@@ -73,63 +73,6 @@
    xgettext.c) with a stream of C tokens.  The comments are
    accumulated in a buffer, and given to xgettext when asked for.  */
 
-enum xgettext_token_type_ty
-{
-  xgettext_token_type_eof,
-  xgettext_token_type_keyword,
-  xgettext_token_type_lparen,
-  xgettext_token_type_rparen,
-  xgettext_token_type_comma,
-  xgettext_token_type_string_literal,
-  xgettext_token_type_symbol
-};
-typedef enum xgettext_token_type_ty xgettext_token_type_ty;
-
-typedef struct xgettext_token_ty xgettext_token_ty;
-struct xgettext_token_ty
-{
-  xgettext_token_type_ty type;
-
-  /* These fields are used only for xgettext_token_type_keyword.  */
-  int argnum1;
-  int argnum2;
-
-  /* This field is used only for xgettext_token_type_string_literal.  */
-  char *string;
-
-  /* These fields are only for
-       xgettext_token_type_keyword,
-       xgettext_token_type_string_literal.  */
-  lex_pos_ty pos;
-};
-
-
-enum token_type_ty
-{
-  token_type_character_constant,
-  token_type_eof,
-  token_type_eoln,
-  token_type_hash,
-  token_type_lparen,
-  token_type_rparen,
-  token_type_comma,
-  token_type_name,
-  token_type_number,
-  token_type_string_literal,
-  token_type_symbol,
-  token_type_white_space
-};
-typedef enum token_type_ty token_type_ty;
-
-typedef struct token_ty token_ty;
-struct token_ty
-{
-  token_type_ty type;
-  char *string;		/* for token_type_name, token_type_string_literal */
-  long number;
-  int line_number;
-};
-
 
 /* ========================= Lexer customization.  ========================= */
 
@@ -147,7 +90,8 @@ x_c_trigraphs ()
 /* If true extract all strings.  */
 static bool extract_all = false;
 
-static hash_table keywords;
+static hash_table c_keywords;
+static hash_table objc_keywords;
 static bool default_keywords = true;
 
 
@@ -158,8 +102,8 @@ x_c_extract_all ()
 }
 
 
-void
-x_c_keyword (const char *name)
+static void
+add_keyword (const char *name, hash_table *keywords)
 {
   if (name == NULL)
     default_keywords = false;
@@ -170,8 +114,8 @@ x_c_keyword (const char *name)
       int argnum2;
       const char *colon;
 
-      if (keywords.table == NULL)
-	init_hash (&keywords, 100);
+      if (keywords->table == NULL)
+	init_hash (keywords, 100);
 
       split_keywordspec (name, &end, &argnum1, &argnum2);
 
@@ -182,19 +126,25 @@ x_c_keyword (const char *name)
 	{
 	  if (argnum1 == 0)
 	    argnum1 = 1;
-	  insert_entry (&keywords, name, end - name,
+	  insert_entry (keywords, name, end - name,
 			(void *) (long) (argnum1 + (argnum2 << 10)));
 	}
     }
 }
 
-bool
-x_c_any_keywords ()
+void
+x_c_keyword (const char *name)
 {
-  return (keywords.filled > 0) || default_keywords;
+  add_keyword (name, &c_keywords);
 }
 
-/* Finish initializing the keywords hash table.
+void
+x_objc_keyword (const char *name)
+{
+  add_keyword (name, &objc_keywords);
+}
+
+/* Finish initializing the keywords hash tables.
    Called after argument processing, before each file is processed.  */
 static void
 init_keywords ()
@@ -208,12 +158,165 @@ init_keywords ()
       x_c_keyword ("dngettext:2,3");
       x_c_keyword ("dcngettext:2,3");
       x_c_keyword ("gettext_noop");
+
+      x_objc_keyword ("gettext");
+      x_objc_keyword ("dgettext:2");
+      x_objc_keyword ("dcgettext:2");
+      x_objc_keyword ("ngettext:1,2");
+      x_objc_keyword ("dngettext:2,3");
+      x_objc_keyword ("dcngettext:2,3");
+      x_objc_keyword ("gettext_noop");
+      x_objc_keyword ("NSLocalizedString");	  /* similar to gettext */
+      x_objc_keyword ("_");			  /* similar to gettext */
+      x_objc_keyword ("NSLocalizedStaticString"); /* similar to gettext_noop */
+      x_objc_keyword ("__");			  /* similar to gettext_noop */
+
       default_keywords = false;
     }
 }
 
+void
+init_flag_table_c ()
+{
+  xgettext_record_flag ("gettext:1:pass-c-format");
+  xgettext_record_flag ("dgettext:2:pass-c-format");
+  xgettext_record_flag ("dcgettext:2:pass-c-format");
+  xgettext_record_flag ("ngettext:1:pass-c-format");
+  xgettext_record_flag ("ngettext:2:pass-c-format");
+  xgettext_record_flag ("dngettext:2:pass-c-format");
+  xgettext_record_flag ("dngettext:3:pass-c-format");
+  xgettext_record_flag ("dcngettext:2:pass-c-format");
+  xgettext_record_flag ("dcngettext:3:pass-c-format");
+  xgettext_record_flag ("gettext_noop:1:pass-c-format");
+  /* <stdio.h> */
+  xgettext_record_flag ("fprintf:2:c-format");
+  xgettext_record_flag ("vfprintf:2:c-format");
+  xgettext_record_flag ("printf:1:c-format");
+  xgettext_record_flag ("vprintf:1:c-format");
+  xgettext_record_flag ("sprintf:2:c-format");
+  xgettext_record_flag ("vsprintf:2:c-format");
+  xgettext_record_flag ("snprintf:3:c-format");
+  xgettext_record_flag ("vsnprintf:3:c-format");
+#if 0 /* These functions are not standard.  */
+  /* <stdio.h> */
+  xgettext_record_flag ("asprintf:2:c-format");
+  xgettext_record_flag ("vasprintf:2:c-format");
+  xgettext_record_flag ("dprintf:2:c-format");
+  xgettext_record_flag ("vdprintf:2:c-format");
+  xgettext_record_flag ("obstack_printf:2:c-format");
+  xgettext_record_flag ("obstack_vprintf:2:c-format");
+  /* <error.h> */
+  xgettext_record_flag ("error:3:c-format");
+  xgettext_record_flag ("error_at_line:5:c-format");
+  /* <argp.h> */
+  xgettext_record_flag ("argp_error:2:c-format");
+  xgettext_record_flag ("argp_failure:2:c-format");
+#endif
+}
 
-/* ================== Reading of characters and tokens.  =================== */
+void
+init_flag_table_objc ()
+{
+  /* Since the settings done in init_flag_table_c() also have an effect for
+     the ObjectiveC parser, we don't have to repeat them here.  */
+  xgettext_record_flag ("gettext:1:pass-objc-format");
+  xgettext_record_flag ("dgettext:2:pass-objc-format");
+  xgettext_record_flag ("dcgettext:2:pass-objc-format");
+  xgettext_record_flag ("ngettext:1:pass-objc-format");
+  xgettext_record_flag ("ngettext:2:pass-objc-format");
+  xgettext_record_flag ("dngettext:2:pass-objc-format");
+  xgettext_record_flag ("dngettext:3:pass-objc-format");
+  xgettext_record_flag ("dcngettext:2:pass-objc-format");
+  xgettext_record_flag ("dcngettext:3:pass-objc-format");
+  xgettext_record_flag ("gettext_noop:1:pass-objc-format");
+  xgettext_record_flag ("NSLocalizedString:1:pass-c-format");
+  xgettext_record_flag ("NSLocalizedString:1:pass-objc-format");
+  xgettext_record_flag ("_:1:pass-c-format");
+  xgettext_record_flag ("_:1:pass-objc-format");
+  xgettext_record_flag ("stringWithFormat::1:objc-format");
+  xgettext_record_flag ("initWithFormat::1:objc-format");
+  xgettext_record_flag ("stringByAppendingFormat::1:objc-format");
+  xgettext_record_flag ("localizedStringWithFormat::1:objc-format");
+  xgettext_record_flag ("appendFormat::1:objc-format");
+}
+
+void
+init_flag_table_gcc_internal ()
+{
+  xgettext_record_flag ("gettext:1:pass-gcc-internal-format");
+  xgettext_record_flag ("dgettext:2:pass-gcc-internal-format");
+  xgettext_record_flag ("dcgettext:2:pass-gcc-internal-format");
+  xgettext_record_flag ("ngettext:1:pass-gcc-internal-format");
+  xgettext_record_flag ("ngettext:2:pass-gcc-internal-format");
+  xgettext_record_flag ("dngettext:2:pass-gcc-internal-format");
+  xgettext_record_flag ("dngettext:3:pass-gcc-internal-format");
+  xgettext_record_flag ("dcngettext:2:pass-gcc-internal-format");
+  xgettext_record_flag ("dcngettext:3:pass-gcc-internal-format");
+  xgettext_record_flag ("gettext_noop:1:pass-gcc-internal-format");
+#if 0 /* This should better be done inside GCC.  */
+  /* grepping for ATTRIBUTE_PRINTF in gcc-3.3/gcc/?*.h */
+  /* c-format.c */
+  xgettext_record_flag ("status_warning:2:gcc-internal-format");
+  /* c-tree.h */
+  xgettext_record_flag ("pedwarn_c99:1:pass-gcc-internal-format");
+  /* collect2.h */
+  //xgettext_record_flag ("error:1:c-format"); // 3 different versions
+  xgettext_record_flag ("notice:1:c-format");
+  //xgettext_record_flag ("fatal:1:c-format"); // 2 different versions
+  xgettext_record_flag ("fatal_perror:1:c-format");
+  /* cpplib.h */
+  xgettext_record_flag ("cpp_error:3:c-format");
+  xgettext_record_flag ("cpp_error_with_line:5:c-format");
+  /* diagnostic.h */
+  xgettext_record_flag ("diagnostic_set_info:2:pass-gcc-internal-format");
+  xgettext_record_flag ("output_printf:2:gcc-internal-format");
+  xgettext_record_flag ("output_verbatim:2:pass-gcc-internal-format");
+  xgettext_record_flag ("verbatim:1:gcc-internal-format");
+  xgettext_record_flag ("inform:1:pass-gcc-internal-format");
+  /* gcc.h */
+  //xgettext_record_flag ("fatal:1:c-format"); // 2 different versions
+  //xgettext_record_flag ("error:1:c-format"); // 3 different versions
+  /* genattrtab.h */
+  xgettext_record_flag ("attr_printf:2:pass-c-format");
+  /* gengtype.h */
+  xgettext_record_flag ("error_at_line:2:pass-c-format");
+  xgettext_record_flag ("xvasprintf:2:pass-c-format");
+  xgettext_record_flag ("xasprintf:1:pass-c-format");
+  xgettext_record_flag ("oprintf:2:pass-c-format");
+  /* gensupport.h */
+  xgettext_record_flag ("message_with_line:2:pass-c-format");
+  /* output.h */
+  xgettext_record_flag ("output_operand_lossage:1:c-format");
+  /* ra.h */
+   xgettext_record_flag ("ra_debug_msg:2:pass-c-format");
+  /* toplev.h */
+  xgettext_record_flag ("fnotice:2:c-format");
+  xgettext_record_flag ("fatal_io_error:2:gcc-internal-format");
+  xgettext_record_flag ("error_for_asm:2:pass-gcc-internal-format");
+  xgettext_record_flag ("warning_for_asm:2:pass-gcc-internal-format");
+  xgettext_record_flag ("error_with_file_and_line:3:pass-gcc-internal-format");
+  xgettext_record_flag ("error_with_decl:2:pass-gcc-internal-format");
+  xgettext_record_flag ("pedwarn:1:gcc-internal-format");
+  xgettext_record_flag ("pedwarn_with_file_and_line:3:gcc-internal-format");
+  xgettext_record_flag ("pedwarn_with_decl:2:gcc-internal-format");
+  xgettext_record_flag ("sorry:1:gcc-internal-format");
+  xgettext_record_flag ("error:1:pass-gcc-internal-format");
+  xgettext_record_flag ("fatal_error:1:pass-gcc-internal-format");
+  xgettext_record_flag ("internal_error:1:pass-gcc-internal-format");
+  xgettext_record_flag ("warning:1:pass-gcc-internal-format");
+  xgettext_record_flag ("warning_with_file_and_line:3:pass-gcc-internal-format");
+  xgettext_record_flag ("warning_with_decl:2:pass-gcc-internal-format");
+  /* f/com.h */
+  xgettext_record_flag ("ffecom_get_invented_identifier:1:pass-c-format");
+  /* f/sts.h */
+  xgettext_record_flag ("ffests_printf:2:pass-c-format");
+  /* java/java-tree.h */
+  xgettext_record_flag ("parse_error_context:2:pass-c-format");
+#endif
+}
+
+
+/* ======================== Reading of characters.  ======================== */
 
 /* Real filename, used in error messages about the input file.  */
 static const char *real_file_name;
@@ -226,10 +329,58 @@ static int line_number;
 static FILE *fp;
 
 
-/* 1. Terminate line by \n, regardless of the external representation of
-   a text line.  Stdio does this for us, we just need to check that
-   there are no I/O errors, and cope with potentially 2 characters of
-   pushback, not just the one that ungetc can cope with.  */
+/* 0. Terminate line by \n, regardless whether the external representation of
+   a line terminator is LF (Unix), CR (Mac) or CR/LF (DOS/Windows).
+   It is debatable whether supporting CR/LF line terminators in C sources
+   on Unix is ISO C or POSIX compliant, but since GCC 3.3 now supports it
+   unconditionally, it must be OK.
+   The so-called "text mode" in stdio on DOS/Windows translates CR/LF to \n
+   automatically, but here we also need this conversion on Unix.  As a side
+   effect, on DOS/Windows we also parse CR/CR/LF into a single \n, but this
+   is not a problem.  */
+
+
+static int
+phase0_getc ()
+{
+  int c;
+
+  c = getc (fp);
+  if (c == EOF)
+    {
+      if (ferror (fp))
+	error (EXIT_FAILURE, errno, _("error while reading \"%s\""),
+	       real_file_name);
+      return EOF;
+    }
+
+  if (c == '\r')
+    {
+      int c1 = getc (fp);
+
+      if (c1 != EOF && c1 != '\n')
+	ungetc (c1, fp);
+
+      /* Seen line terminator CR or CR/LF.  */
+      return '\n';
+    }
+
+  return c;
+}
+
+
+/* Only one pushback character supported, and not '\n'.  */
+static inline void
+phase0_ungetc (int c)
+{
+  if (c != EOF)
+    ungetc (c, fp);
+}
+
+
+/* 1. line_number handling.  Combine backslash-newline to nothing.
+   Cope with potentially 2 characters of pushback, not just the one that
+   ungetc can cope with.  */
 
 /* Maximum used guaranteed to be < 4.  */
 static unsigned char phase1_pushback[4];
@@ -250,33 +401,18 @@ phase1_getc ()
     }
   for (;;)
     {
-      c = getc (fp);
+      c = phase0_getc ();
       switch (c)
 	{
-	case EOF:
-	  if (ferror (fp))
-	    {
-	    bomb:
-	      error (EXIT_FAILURE, errno, _("\
-error while reading \"%s\""), real_file_name);
-	    }
-	  return EOF;
-
 	case '\n':
 	  ++line_number;
 	  return '\n';
 
 	case '\\':
-	  c = getc (fp);
-	  if (c == EOF)
-	    {
-	      if (ferror (fp))
-		goto bomb;
-	      return '\\';
-	    }
+	  c = phase0_getc (fp);
 	  if (c != '\n')
 	    {
-	      ungetc (c, fp);
+	      phase0_ungetc (c);
 	      return '\\';
 	    }
 	  ++line_number;
@@ -426,7 +562,7 @@ comment_add (int c)
 {
   if (buflen >= bufmax)
     {
-      bufmax += 100;
+      bufmax = 2 * bufmax + 10;
       buffer = xrealloc (buffer, bufmax);
     }
   buffer[buflen++] = c;
@@ -441,7 +577,7 @@ comment_line_end (size_t chars_to_remove)
     --buflen;
   if (chars_to_remove == 0 && buflen >= bufmax)
     {
-      bufmax += 100;
+      bufmax = 2 * bufmax + 10;
       buffer = xrealloc (buffer, bufmax);
     }
   buffer[buflen] = '\0';
@@ -526,7 +662,9 @@ phase4_getc ()
 	  c = phase3_getc ();
 	  if (c == '\n' || c == EOF)
 	    break;
-	  comment_add (c);
+	  /* We skip all leading white space, but not EOLs.  */
+	  if (!(buflen == 0 && (c == ' ' || c == '\t')))
+	    comment_add (c);
 	}
       comment_line_end (0);
       last_comment_line = newline_count;
@@ -542,6 +680,41 @@ phase4_ungetc (int c)
 }
 
 
+/* ========================== Reading of tokens.  ========================== */
+
+
+/* True if ObjectiveC extensions are recognized.  */
+static bool objc_extensions;
+
+enum token_type_ty
+{
+  token_type_character_constant,	/* 'x' */
+  token_type_eof,
+  token_type_eoln,
+  token_type_hash,			/* # */
+  token_type_lparen,			/* ( */
+  token_type_rparen,			/* ) */
+  token_type_comma,			/* , */
+  token_type_colon,			/* : */
+  token_type_name,			/* abc */
+  token_type_number,			/* 2.7 */
+  token_type_string_literal,		/* "abc" */
+  token_type_symbol,			/* < > = etc. */
+  token_type_objc_special,		/* @ */
+  token_type_white_space
+};
+typedef enum token_type_ty token_type_ty;
+
+typedef struct token_ty token_ty;
+struct token_ty
+{
+  token_type_ty type;
+  char *string;		/* for token_type_name, token_type_string_literal */
+  long number;
+  int line_number;
+};
+
+
 /* 7. Replace escape sequences within character strings with their
    single character equivalents.  This is called from phase 5, because
    we don't have to worry about the #include argument.  There are
@@ -607,8 +780,8 @@ phase7_getc ()
       return '\b';
 
       /* The \e escape is preculiar to gcc, and assumes an ASCII
-         character set (or superset).  We don't provide support for it
-         here.  */
+	 character set (or superset).  We don't provide support for it
+	 here.  */
 
     case 'f':
       return '\f';
@@ -774,7 +947,7 @@ phase5_get (token_ty *tp)
 	{
 	  if (bufpos >= bufmax)
 	    {
-	      bufmax += 100;
+	      bufmax = 2 * bufmax + 10;
 	      buffer = xrealloc (buffer, bufmax);
 	    }
 	  buffer[bufpos++] = c;
@@ -804,7 +977,7 @@ phase5_get (token_ty *tp)
 	}
       if (bufpos >= bufmax)
 	{
-	  bufmax += 100;
+	  bufmax = 2 * bufmax + 10;
 	  buffer = xrealloc (buffer, bufmax);
 	}
       buffer[bufpos] = 0;
@@ -838,7 +1011,7 @@ phase5_get (token_ty *tp)
 	{
 	  if (bufpos >= bufmax)
 	    {
-	      bufmax += 100;
+	      bufmax = 2 * bufmax + 10;
 	      buffer = xrealloc (buffer, bufmax);
 	    }
 	  buffer[bufpos++] = c;
@@ -849,7 +1022,7 @@ phase5_get (token_ty *tp)
 	    case 'E':
 	      if (bufpos >= bufmax)
 		{
-		  bufmax += 100;
+		  bufmax = 2 * bufmax + 10;
 		  buffer = xrealloc (buffer, bufmax);
 		}
 	      buffer[bufpos++] = c;
@@ -884,7 +1057,7 @@ phase5_get (token_ty *tp)
 	}
       if (bufpos >= bufmax)
 	{
-	  bufmax += 100;
+	  bufmax = 2 * bufmax + 10;
 	  buffer = xrealloc (buffer, bufmax);
 	}
       buffer[bufpos] = 0;
@@ -939,14 +1112,14 @@ phase5_get (token_ty *tp)
 	    c = '\'';
 	  if (bufpos >= bufmax)
 	    {
-	      bufmax += 100;
+	      bufmax = 2 * bufmax + 10;
 	      buffer = xrealloc (buffer, bufmax);
 	    }
 	  buffer[bufpos++] = c;
 	}
       if (bufpos >= bufmax)
 	{
-	  bufmax += 100;
+	  bufmax = 2 * bufmax + 10;
 	  buffer = xrealloc (buffer, bufmax);
 	}
       buffer[bufpos] = 0;
@@ -970,10 +1143,22 @@ phase5_get (token_ty *tp)
       tp->type = token_type_hash;
       return;
 
+    case ':':
+      tp->type = token_type_colon;
+      return;
+
+    case '@':
+      if (objc_extensions)
+	{
+	  tp->type = token_type_objc_special;
+	  return;
+	}
+      /* FALLTHROUGH */
+
     default:
       /* We could carefully recognize each of the 2 and 3 character
-        operators, but it is not necessary, as we only need to recognize
-        gettext invocations.  Don't bother.  */
+	operators, but it is not necessary, as we only need to recognize
+	gettext invocations.  Don't bother.  */
       tp->type = token_type_symbol;
       return;
     }
@@ -1085,7 +1270,7 @@ phase6_get (token_ty *tp)
 	      /* Accumulate.  */
 	      if (bufpos >= bufmax)
 		{
-		  bufmax += 100;
+		  bufmax = 2 * bufmax + 10;
 		  buf = xrealloc (buf, bufmax * sizeof (buf[0]));
 		}
 	      buf[bufpos++] = *tp;
@@ -1188,13 +1373,76 @@ phase8a_get (token_ty *tp)
     }
 }
 
-static void
+static inline void
 phase8a_unget (token_ty *tp)
 {
   phase6_unget (tp);
 }
 
 
+/* 8b. Drop whitespace.  */
+static void
+phase8b_get (token_ty *tp)
+{
+  for (;;)
+    {
+      phase8a_get (tp);
+
+      if (tp->type == token_type_white_space)
+	continue;
+      if (tp->type == token_type_eoln)
+	{
+	  /* We have to track the last occurrence of a string.  One
+	     mode of xgettext allows to group an extracted message
+	     with a comment for documentation.  The rule which states
+	     which comment is assumed to be grouped with the message
+	     says it should immediately precede it.  Our
+	     interpretation: between the last line of the comment and
+	     the line in which the keyword is found must be no line
+	     with non-white space tokens.  */
+	  ++newline_count;
+	  if (last_non_comment_line > last_comment_line)
+	    xgettext_comment_reset ();
+	  continue;
+	}
+      break;
+    }
+}
+
+static inline void
+phase8b_unget (token_ty *tp)
+{
+  phase8a_unget (tp);
+}
+
+
+/* 8c. In ObjectiveC mode, drop '@' before a literal string.  We need to
+   do this before performing concatenation of adjacent string literals.  */
+static void
+phase8c_get (token_ty *tp)
+{
+  token_ty tmp;
+
+  phase8b_get (tp);
+  if (tp->type != token_type_objc_special)
+    return;
+  phase8b_get (&tmp);
+  if (tmp.type != token_type_string_literal)
+    {
+      phase8b_unget (&tmp);
+      return;
+    }
+  /* Drop the '@' token and return immediately the following string.  */
+  *tp = tmp;
+}
+
+static inline void
+phase8c_unget (token_ty *tp)
+{
+  phase8b_unget (tp);
+}
+
+
 /* 8. Concatenate adjacent string literals to form single string
    literals (because we don't expand macros, there are a few things we
    will miss).  */
@@ -1202,7 +1450,7 @@ phase8a_unget (token_ty *tp)
 static void
 phase8_get (token_ty *tp)
 {
-  phase8a_get (tp);
+  phase8c_get (tp);
   if (tp->type != token_type_string_literal)
     return;
   for (;;)
@@ -1210,14 +1458,10 @@ phase8_get (token_ty *tp)
       token_ty tmp;
       size_t len;
 
-      phase8a_get (&tmp);
-      if (tmp.type == token_type_white_space)
-	continue;
-      if (tmp.type == token_type_eoln)
-	continue;
+      phase8c_get (&tmp);
       if (tmp.type != token_type_string_literal)
 	{
-	  phase8a_unget (&tmp);
+	  phase8c_unget (&tmp);
 	  return;
 	}
       len = strlen (tp->string);
@@ -1228,6 +1472,43 @@ phase8_get (token_ty *tp)
 }
 
 
+/* ===================== Reading of high-level tokens.  ==================== */
+
+
+enum xgettext_token_type_ty
+{
+  xgettext_token_type_eof,
+  xgettext_token_type_keyword,
+  xgettext_token_type_symbol,
+  xgettext_token_type_lparen,
+  xgettext_token_type_rparen,
+  xgettext_token_type_comma,
+  xgettext_token_type_colon,
+  xgettext_token_type_string_literal,
+  xgettext_token_type_other
+};
+typedef enum xgettext_token_type_ty xgettext_token_type_ty;
+
+typedef struct xgettext_token_ty xgettext_token_ty;
+struct xgettext_token_ty
+{
+  xgettext_token_type_ty type;
+
+  /* These fields are used only for xgettext_token_type_keyword.  */
+  int argnum1;
+  int argnum2;
+
+  /* This field is used only for xgettext_token_type_string_literal,
+     xgettext_token_type_keyword, xgettext_token_type_symbol.  */
+  char *string;
+
+  /* These fields are only for
+       xgettext_token_type_keyword,
+       xgettext_token_type_string_literal.  */
+  lex_pos_ty pos;
+};
+
+
 /* 9. Convert the remaining preprocessing tokens to C tokens and
    discards any white space from the translation unit.  */
 
@@ -1246,28 +1527,11 @@ x_c_lex (xgettext_token_ty *tp)
 	  tp->type = xgettext_token_type_eof;
 	  return;
 
-	case token_type_white_space:
-	  break;
-
-	case token_type_eoln:
-	  /* We have to track the last occurrence of a string.  One
-	     mode of xgettext allows to group an extracted message
-	     with a comment for documentation.  The rule which states
-	     which comment is assumed to be grouped with the message
-	     says it should immediately precede it.  Our
-	     interpretation: between the last line of the comment and
-	     the line in which the keyword is found must be no line
-	     with non-white space tokens.  */
-	  ++newline_count;
-	  if (last_non_comment_line > last_comment_line)
-	    xgettext_comment_reset ();
-	  break;
-
 	case token_type_name:
 	  last_non_comment_line = newline_count;
 
-	  if (find_entry (&keywords, token.string, strlen (token.string),
-			  &keyword_value)
+	  if (find_entry (objc_extensions ? &objc_keywords : &c_keywords,
+			  token.string, strlen (token.string), &keyword_value)
 	      == 0)
 	    {
 	      tp->type = xgettext_token_type_keyword;
@@ -1278,7 +1542,7 @@ x_c_lex (xgettext_token_ty *tp)
 	    }
 	  else
 	    tp->type = xgettext_token_type_symbol;
-	  free (token.string);
+	  tp->string = token.string;
 	  return;
 
 	case token_type_lparen:
@@ -1299,6 +1563,12 @@ x_c_lex (xgettext_token_ty *tp)
 	  tp->type = xgettext_token_type_comma;
 	  return;
 
+	case token_type_colon:
+	  last_non_comment_line = newline_count;
+
+	  tp->type = xgettext_token_type_colon;
+	  return;
+
 	case token_type_string_literal:
 	  last_non_comment_line = newline_count;
 
@@ -1311,7 +1581,7 @@ x_c_lex (xgettext_token_ty *tp)
 	default:
 	  last_non_comment_line = newline_count;
 
-	  tp->type = xgettext_token_type_symbol;
+	  tp->type = xgettext_token_type_other;
 	  return;
 	}
     }
@@ -1320,6 +1590,11 @@ x_c_lex (xgettext_token_ty *tp)
 
 /* ========================= Extracting strings.  ========================== */
 
+
+/* Context lookup table.  */
+static flag_context_list_table_ty *flag_context_list_table;
+
+
 /* The file is broken into tokens.  Scan the token stream, looking for
    a keyword, followed by a left paren, followed by a string.  When we
    see this sequence, we have something to remember.  We assume we are
@@ -1344,6 +1619,8 @@ x_c_lex (xgettext_token_ty *tp)
    Return true upon eof, false upon closing parenthesis.  */
 static bool
 extract_parenthesized (message_list_ty *mlp,
+		       flag_context_ty outer_context,
+		       flag_context_list_iterator_ty context_iter,
 		       int commas_to_skip, int plural_commas)
 {
   /* Remember the message containing the msgid, for msgid_plural.  */
@@ -1354,6 +1631,17 @@ extract_parenthesized (message_list_ty *mlp,
   /* Parameters of the keyword just seen.  Defined only in state 1.  */
   int next_commas_to_skip = -1;
   int next_plural_commas = 0;
+  /* Context iterator that will be used if the next token is a '('.  */
+  flag_context_list_iterator_ty next_context_iter =
+    passthrough_context_list_iterator;
+  /* Context iterator that will be used if the next token is a ':'.
+     (Objective C selector syntax.)  */
+  flag_context_list_iterator_ty selectorcall_context_iter =
+    passthrough_context_list_iterator;
+  /* Current context.  */
+  flag_context_ty inner_context =
+    inherited_context (outer_context,
+		       flag_context_list_iterator_advance (&context_iter));
 
   /* Start state is 0.  */
   state = 0;
@@ -1366,37 +1654,49 @@ extract_parenthesized (message_list_ty *mlp,
       switch (token.type)
 	{
 	case xgettext_token_type_keyword:
-	  /* No need to bother if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
 	  next_commas_to_skip = token.argnum1 - 1;
 	  next_plural_commas = (token.argnum2 > token.argnum1
 				? token.argnum2 - token.argnum1 : 0);
 	  state = 1;
+	  goto keyword_or_symbol;
+
+	case xgettext_token_type_symbol:
+	  state = 0;
+	keyword_or_symbol:
+	  next_context_iter =
+	    flag_context_list_iterator (
+	      flag_context_list_table_lookup (
+		flag_context_list_table,
+		token.string, strlen (token.string)));
+	  if (objc_extensions)
+	    {
+	      size_t token_string_len = strlen (token.string);
+	      token.string = xrealloc (token.string, token_string_len + 2);
+	      token.string[token_string_len] = ':';
+	      token.string[token_string_len + 1] = '\0';
+	      selectorcall_context_iter =
+		flag_context_list_iterator (
+		  flag_context_list_table_lookup (
+		    flag_context_list_table,
+		    token.string, token_string_len + 1));
+	    }
+	  free (token.string);
 	  continue;
 
 	case xgettext_token_type_lparen:
-	  /* No need to recurse if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
-	  if (state
-	      ?  extract_parenthesized (mlp, next_commas_to_skip,
-					next_plural_commas)
-	      : extract_parenthesized (mlp, -1, 0))
+	  if (extract_parenthesized (mlp, inner_context, next_context_iter,
+				     state ? next_commas_to_skip : -1,
+				     state ? next_plural_commas : 0))
 	    return true;
+	  next_context_iter = null_context_list_iterator;
+	  selectorcall_context_iter = null_context_list_iterator;
 	  state = 0;
 	  continue;
 
 	case xgettext_token_type_rparen:
-	  /* No need to return if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
 	  return false;
 
 	case xgettext_token_type_comma:
-	  /* No need to bother if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
 	  if (commas_to_skip >= 0)
 	    {
 	      if (commas_to_skip > 0)
@@ -1410,12 +1710,37 @@ extract_parenthesized (message_list_ty *mlp,
 		else
 		  commas_to_skip = -1;
 	    }
+	  inner_context =
+	    inherited_context (outer_context,
+			       flag_context_list_iterator_advance (
+				 &context_iter));
+	  next_context_iter = passthrough_context_list_iterator;
+	  selectorcall_context_iter = passthrough_context_list_iterator;
+	  state = 0;
+	  continue;
+
+	case xgettext_token_type_colon:
+	  if (objc_extensions)
+	    {
+	      context_iter = selectorcall_context_iter;
+	      inner_context =
+		inherited_context (inner_context,
+				   flag_context_list_iterator_advance (
+				     &context_iter));
+	      next_context_iter = passthrough_context_list_iterator;
+	      selectorcall_context_iter = passthrough_context_list_iterator;
+	    }
+	  else
+	    {
+	      next_context_iter = null_context_list_iterator;
+	      selectorcall_context_iter = null_context_list_iterator;
+	    }
 	  state = 0;
 	  continue;
 
 	case xgettext_token_type_string_literal:
 	  if (extract_all)
-	    remember_a_message (mlp, token.string, &token.pos);
+	    remember_a_message (mlp, token.string, inner_context, &token.pos);
 	  else
 	    {
 	      if (commas_to_skip == 0)
@@ -1423,8 +1748,9 @@ extract_parenthesized (message_list_ty *mlp,
 		  if (plural_mp == NULL)
 		    {
 		      /* Seen an msgid.  */
-		      message_ty *mp = remember_a_message (mlp, token.string,
-							   &token.pos);
+		      message_ty *mp =
+			remember_a_message (mlp, token.string,
+					    inner_context, &token.pos);
 		      if (plural_commas > 0)
 			plural_mp = mp;
 		    }
@@ -1432,17 +1758,21 @@ extract_parenthesized (message_list_ty *mlp,
 		    {
 		      /* Seen an msgid_plural.  */
 		      remember_a_message_plural (plural_mp, token.string,
-						 &token.pos);
+						 inner_context, &token.pos);
 		      plural_mp = NULL;
 		    }
 		}
 	      else
 		free (token.string);
-	      state = 0;
 	    }
+	  next_context_iter = null_context_list_iterator;
+	  selectorcall_context_iter = null_context_list_iterator;
+	  state = 0;
 	  continue;
 
-	case xgettext_token_type_symbol:
+	case xgettext_token_type_other:
+	  next_context_iter = null_context_list_iterator;
+	  selectorcall_context_iter = null_context_list_iterator;
 	  state = 0;
 	  continue;
 
@@ -1456,10 +1786,11 @@ extract_parenthesized (message_list_ty *mlp,
 }
 
 
-void
-extract_c (FILE *f,
-	   const char *real_filename, const char *logical_filename,
-	   msgdomain_list_ty *mdlp)
+static void
+extract_whole_file (FILE *f,
+		    const char *real_filename, const char *logical_filename,
+		    flag_context_list_table_ty *flag_table,
+		    msgdomain_list_ty *mdlp)
 {
   message_list_ty *mlp = mdlp->item[0]->messages;
 
@@ -1472,11 +1803,14 @@ extract_c (FILE *f,
   last_comment_line = -1;
   last_non_comment_line = -1;
 
+  flag_context_list_table = flag_table;
+
   init_keywords ();
 
   /* Eat tokens until eof is seen.  When extract_parenthesized returns
      due to an unbalanced closing parenthesis, just restart it.  */
-  while (!extract_parenthesized (mlp, -1, 0))
+  while (!extract_parenthesized (mlp, null_context, null_context_list_iterator,
+				 -1, 0))
     ;
 
   /* Close scanner.  */
@@ -1485,3 +1819,24 @@ extract_c (FILE *f,
   logical_file_name = NULL;
   line_number = 0;
 }
+
+
+void
+extract_c (FILE *f,
+	   const char *real_filename, const char *logical_filename,
+	   flag_context_list_table_ty *flag_table,
+	   msgdomain_list_ty *mdlp)
+{
+  objc_extensions = false;
+  extract_whole_file (f, real_filename, logical_filename, flag_table, mdlp);
+}
+
+void
+extract_objc (FILE *f,
+	      const char *real_filename, const char *logical_filename,
+	      flag_context_list_table_ty *flag_table,
+	      msgdomain_list_ty *mdlp)
+{
+  objc_extensions = true;
+  extract_whole_file (f, real_filename, logical_filename, flag_table, mdlp);
+}
