@@ -1,5 +1,5 @@
 /* xgettext glade backend.
-   Copyright (C) 2002 Free Software Foundation, Inc.
+   Copyright (C) 2002-2003 Free Software Foundation, Inc.
 
    This file was written by Bruno Haible <haible@clisp.cons.org>, 2002.
 
@@ -35,18 +35,16 @@
 #endif
 
 #include "message.h"
-#include "x-glade.h"
 #include "xgettext.h"
+#include "x-glade.h"
 #include "error.h"
 #include "xerror.h"
 #include "basename.h"
 #include "progname.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "exit.h"
 #include "hash.h"
 #include "po-charset.h"
-#include "msgl-ascii.h"
-#include "msgl-iconv.h"
 #include "gettext.h"
 
 #define _(s) gettext(s)
@@ -232,8 +230,52 @@ start_element_handler (void *userData, const char *name,
   stack[stack_depth - 1].extract_string = false;
 
   p = &stack[stack_depth];
-  p->extract_string =
-    (find_entry (&keywords, name, strlen (name), &hash_result) == 0);
+  p->extract_string = extract_all;
+  /* In Glade 1, a few specific elements are translatable.  */
+  if (!p->extract_string)
+    p->extract_string =
+      (find_entry (&keywords, name, strlen (name), &hash_result) == 0);
+  /* In Glade 2, all <property> and <atkproperty> elements are translatable
+     that have the attribute translatable="yes".  */
+  if (!p->extract_string
+      && (strcmp (name, "property") == 0 || strcmp (name, "atkproperty") == 0))
+    {
+      bool has_translatable = false;
+      const char **attp = attributes;
+      while (*attp != NULL)
+	{
+	  if (strcmp (attp[0], "translatable") == 0)
+	    {
+	      has_translatable = (strcmp (attp[1], "yes") == 0);
+	      break;
+	    }
+	  attp += 2;
+	}
+      p->extract_string = has_translatable;
+    }
+  if (!p->extract_string
+      && strcmp (name, "atkaction") == 0)
+    {
+      const char **attp = attributes;
+      while (*attp != NULL)
+	{
+	  if (strcmp (attp[0], "description") == 0)
+	    {
+	      if (strcmp (attp[1], "") != 0)
+		{
+		  lex_pos_ty pos;
+
+		  pos.file_name = logical_file_name;
+		  pos.line_number = XML_GetCurrentLineNumber (parser);
+
+		  remember_a_message (mlp, xstrdup (attp[1]),
+				      null_context, &pos);
+		}
+	      break;
+	    }
+	  attp += 2;
+	}
+    }
   p->lineno = XML_GetCurrentLineNumber (parser);
   p->buffer = NULL;
   p->bufmax = 0;
@@ -263,7 +305,7 @@ end_element_handler (void *userData, const char *name)
 	  pos.file_name = logical_file_name;
 	  pos.line_number = p->lineno;
 
-	  remember_a_message (mlp, p->buffer, &pos);
+	  remember_a_message (mlp, p->buffer, null_context, &pos);
 	  p->buffer = NULL;
 	}
     }
@@ -351,6 +393,8 @@ do_extract_glade (FILE *fp,
   XML_SetCharacterDataHandler (parser, character_data_handler);
   XML_SetCommentHandler (parser, comment_handler);
 
+  stack_depth = 0;
+
   while (!feof (fp))
     {
       char buf[4096];
@@ -390,6 +434,7 @@ error while reading \"%s\""), real_filename);
 void
 extract_glade (FILE *fp,
 	       const char *real_filename, const char *logical_filename,
+	       flag_context_list_table_ty *flag_table,
 	       msgdomain_list_ty *mdlp)
 {
 #if DYNLOAD_LIBEXPAT || HAVE_LIBEXPAT
