@@ -1,5 +1,5 @@
 /* xgettext PHP backend.
-   Copyright (C) 2001-2002 Free Software Foundation, Inc.
+   Copyright (C) 2001-2003 Free Software Foundation, Inc.
 
    This file was written by Bruno Haible <bruno@clisp.org>, 2002.
 
@@ -27,10 +27,10 @@
 #include <stdlib.h>
 
 #include "message.h"
-#include "x-php.h"
 #include "xgettext.h"
+#include "x-php.h"
 #include "error.h"
-#include "xmalloc.h"
+#include "xalloc.h"
 #include "exit.h"
 #include "gettext.h"
 
@@ -40,26 +40,6 @@
 /* The PHP syntax is defined in phpdoc/manual/langref.html.
    See also php-4.1.0/Zend/zend_language_scanner.l.  */
 
-enum token_type_ty
-{
-  token_type_eof,
-  token_type_lparen,		/* ( */
-  token_type_rparen,		/* ) */
-  token_type_comma,		/* , */
-  token_type_string_literal,	/* "abc" */
-  token_type_symbol,		/* symbol, number */
-  token_type_other		/* misc. operator */
-};
-typedef enum token_type_ty token_type_ty;
-
-typedef struct token_ty token_ty;
-struct token_ty
-{
-  token_type_ty type;
-  char *string;		/* for token_type_string_literal, token_type_symbol */
-  int line_number;
-};
-
 
 /* ====================== Keyword set customization.  ====================== */
 
@@ -118,10 +98,31 @@ init_keywords ()
       x_php_keyword ("gettext");
       x_php_keyword ("dgettext:2");
       x_php_keyword ("dcgettext:2");
+      /* The following were added in PHP 4.2.0.  */
+      x_php_keyword ("ngettext:1,2");
+      x_php_keyword ("dngettext:2,3");
+      x_php_keyword ("dcngettext:2,3");
       default_keywords = false;
     }
 }
 
+void
+init_flag_table_php ()
+{
+  xgettext_record_flag ("_:1:pass-php-format");
+  xgettext_record_flag ("gettext:1:pass-php-format");
+  xgettext_record_flag ("dgettext:2:pass-php-format");
+  xgettext_record_flag ("dcgettext:2:pass-php-format");
+  xgettext_record_flag ("ngettext:1:pass-php-format");
+  xgettext_record_flag ("ngettext:2:pass-php-format");
+  xgettext_record_flag ("dngettext:2:pass-php-format");
+  xgettext_record_flag ("dngettext:3:pass-php-format");
+  xgettext_record_flag ("dcngettext:2:pass-php-format");
+  xgettext_record_flag ("dcngettext:3:pass-php-format");
+  xgettext_record_flag ("sprintf:1:php-format");
+  xgettext_record_flag ("printf:1:php-format");
+}
+
 
 /* ======================== Reading of characters.  ======================== */
 
@@ -537,7 +538,7 @@ comment_add (int c)
 {
   if (buflen >= bufmax)
     {
-      bufmax += 100;
+      bufmax = 2 * bufmax + 10;
       buffer = xrealloc (buffer, bufmax);
     }
   buffer[buflen++] = c;
@@ -552,7 +553,7 @@ comment_line_end (size_t chars_to_remove)
     --buflen;
   if (chars_to_remove == 0 && buflen >= bufmax)
     {
-      bufmax += 100;
+      bufmax = 2 * bufmax + 10;
       buffer = xrealloc (buffer, bufmax);
     }
   buffer[buflen] = '\0';
@@ -705,12 +706,38 @@ phase3_getc ()
     return c;
 }
 
+#ifdef unused
 static void
 phase3_ungetc (int c)
 {
   if (c != EOF)
     phase3_pushback[phase3_pushback_length++] = c;
 }
+#endif
+
+
+/* ========================== Reading of tokens.  ========================== */
+
+
+enum token_type_ty
+{
+  token_type_eof,
+  token_type_lparen,		/* ( */
+  token_type_rparen,		/* ) */
+  token_type_comma,		/* , */
+  token_type_string_literal,	/* "abc" */
+  token_type_symbol,		/* symbol, number */
+  token_type_other		/* misc. operator */
+};
+typedef enum token_type_ty token_type_ty;
+
+typedef struct token_ty token_ty;
+struct token_ty
+{
+  token_type_ty type;
+  char *string;		/* for token_type_string_literal, token_type_symbol */
+  int line_number;
+};
 
 
 /* Free the memory pointed to by a 'struct token_ty'.  */
@@ -773,7 +800,7 @@ x_php_lex (token_ty *tp)
 	    {
 	      if (bufpos >= bufmax)
 		{
-		  bufmax += 100;
+		  bufmax = 2 * bufmax + 10;
 		  buffer = xrealloc (buffer, bufmax);
 		}
 	      buffer[bufpos++] = c;
@@ -803,7 +830,7 @@ x_php_lex (token_ty *tp)
 	    }
 	  if (bufpos >= bufmax)
 	    {
-	      bufmax += 100;
+	      bufmax = 2 * bufmax + 10;
 	      buffer = xrealloc (buffer, bufmax);
 	    }
 	  buffer[bufpos] = 0;
@@ -830,14 +857,14 @@ x_php_lex (token_ty *tp)
 		}
 	      if (bufpos >= bufmax)
 		{
-		  bufmax += 100;
+		  bufmax = 2 * bufmax + 10;
 		  buffer = xrealloc (buffer, bufmax);
 		}
 	      buffer[bufpos++] = c;
 	    }
 	  if (bufpos >= bufmax)
 	    {
-	      bufmax += 100;
+	      bufmax = 2 * bufmax + 10;
 	      buffer = xrealloc (buffer, bufmax);
 	    }
 	  buffer[bufpos] = 0;
@@ -967,14 +994,14 @@ x_php_lex (token_ty *tp)
 		}
 	      if (bufpos >= bufmax)
 		{
-		  bufmax += 100;
+		  bufmax = 2 * bufmax + 10;
 		  buffer = xrealloc (buffer, bufmax);
 		}
 	      buffer[bufpos++] = c;
 	    }
 	  if (bufpos >= bufmax)
 	    {
-	      bufmax += 100;
+	      bufmax = 2 * bufmax + 10;
 	      buffer = xrealloc (buffer, bufmax);
 	    }
 	  buffer[bufpos] = 0;
@@ -1029,7 +1056,7 @@ x_php_lex (token_ty *tp)
 		      {
 			if (bufpos >= bufmax)
 			  {
-			    bufmax += 100;
+			    bufmax = 2 * bufmax + 10;
 			    buffer = xrealloc (buffer, bufmax);
 			  }
 			buffer[bufpos++] = c;
@@ -1152,8 +1179,14 @@ x_php_lex (token_ty *tp)
     }
 }
 
+
 /* ========================= Extracting strings.  ========================== */
 
+
+/* Context lookup table.  */
+static flag_context_list_table_ty *flag_context_list_table;
+
+
 /* The file is broken into tokens.  Scan the token stream, looking for
    a keyword, followed by a left paren, followed by a string.  When we
    see this sequence, we have something to remember.  We assume we are
@@ -1178,6 +1211,8 @@ x_php_lex (token_ty *tp)
    Return true upon eof, false upon closing parenthesis.  */
 static bool
 extract_parenthesized (message_list_ty *mlp,
+		       flag_context_ty outer_context,
+		       flag_context_list_iterator_ty context_iter,
 		       int commas_to_skip, int plural_commas)
 {
   /* Remember the message containing the msgid, for msgid_plural.  */
@@ -1188,6 +1223,13 @@ extract_parenthesized (message_list_ty *mlp,
   /* Parameters of the keyword just seen.  Defined only in state 1.  */
   int next_commas_to_skip = -1;
   int next_plural_commas = 0;
+  /* Context iterator that will be used if the next token is a '('.  */
+  flag_context_list_iterator_ty next_context_iter =
+    passthrough_context_list_iterator;
+  /* Current context.  */
+  flag_context_ty inner_context =
+    inherited_context (outer_context,
+		       flag_context_list_iterator_advance (&context_iter));
 
   /* Start state is 0.  */
   state = 0;
@@ -1200,50 +1242,44 @@ extract_parenthesized (message_list_ty *mlp,
       switch (token.type)
 	{
 	case token_type_symbol:
-	  /* No need to bother if we extract all strings anyway.  */
-	  if (!extract_all)
-	    {
-	      void *keyword_value;
+	  {
+	    void *keyword_value;
 
-	      if (find_entry (&keywords, token.string, strlen (token.string),
-			      &keyword_value)
-		  == 0)
-		{
-		  int argnum1 = (int) (long) keyword_value & ((1 << 10) - 1);
-		  int argnum2 = (int) (long) keyword_value >> 10;
+	    if (find_entry (&keywords, token.string, strlen (token.string),
+			    &keyword_value)
+		== 0)
+	      {
+		int argnum1 = (int) (long) keyword_value & ((1 << 10) - 1);
+		int argnum2 = (int) (long) keyword_value >> 10;
 
-		  next_commas_to_skip = argnum1 - 1;
-		  next_plural_commas = (argnum2 > argnum1 ? argnum2 - argnum1 : 0);
-		  state = 1;
-		}
-	      else
-		state = 0;
-	    }
+		next_commas_to_skip = argnum1 - 1;
+		next_plural_commas = (argnum2 > argnum1 ? argnum2 - argnum1 : 0);
+		state = 1;
+	      }
+	    else
+	      state = 0;
+	  }
+	  next_context_iter =
+	    flag_context_list_iterator (
+	      flag_context_list_table_lookup (
+		flag_context_list_table,
+		token.string, strlen (token.string)));
 	  free (token.string);
 	  continue;
 
 	case token_type_lparen:
-	  /* No need to recurse if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
-	  if (state
-	      ?  extract_parenthesized (mlp, next_commas_to_skip,
-					next_plural_commas)
-	      : extract_parenthesized (mlp, -1, 0))
+	  if (extract_parenthesized (mlp, inner_context, next_context_iter,
+				     state ? next_commas_to_skip : -1,
+				     state ? next_plural_commas: 0))
 	    return true;
+	  next_context_iter = null_context_list_iterator;
 	  state = 0;
 	  continue;
 
 	case token_type_rparen:
-	  /* No need to return if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
 	  return false;
 
 	case token_type_comma:
-	  /* No need to bother if we extract all strings anyway.  */
-	  if (extract_all)
-	    continue;
 	  if (commas_to_skip >= 0)
 	    {
 	      if (commas_to_skip > 0)
@@ -1257,6 +1293,11 @@ extract_parenthesized (message_list_ty *mlp,
 		else
 		  commas_to_skip = -1;
 	    }
+	  inner_context =
+	    inherited_context (outer_context,
+			       flag_context_list_iterator_advance (
+				 &context_iter));
+	  next_context_iter = passthrough_context_list_iterator;
 	  state = 0;
 	  continue;
 
@@ -1267,7 +1308,7 @@ extract_parenthesized (message_list_ty *mlp,
 	    pos.line_number = token.line_number;
 
 	    if (extract_all)
-	      remember_a_message (mlp, token.string, &pos);
+	      remember_a_message (mlp, token.string, inner_context, &pos);
 	    else
 	      {
 		if (commas_to_skip == 0)
@@ -1275,8 +1316,9 @@ extract_parenthesized (message_list_ty *mlp,
 		    if (plural_mp == NULL)
 		      {
 			/* Seen an msgid.  */
-			message_ty *mp = remember_a_message (mlp, token.string,
-							     &pos);
+			message_ty *mp =
+			  remember_a_message (mlp, token.string,
+					      inner_context, &pos);
 			if (plural_commas > 0)
 			  plural_mp = mp;
 		      }
@@ -1284,18 +1326,20 @@ extract_parenthesized (message_list_ty *mlp,
 		      {
 			/* Seen an msgid_plural.  */
 			remember_a_message_plural (plural_mp, token.string,
-						   &pos);
+						   inner_context, &pos);
 			plural_mp = NULL;
 		      }
 		  }
 		else
 		  free (token.string);
-		state = 0;
 	      }
-	    continue;
 	  }
+	  next_context_iter = null_context_list_iterator;
+	  state = 0;
+	  continue;
 
 	case token_type_other:
+	  next_context_iter = null_context_list_iterator;
 	  state = 0;
 	  continue;
 
@@ -1312,6 +1356,7 @@ extract_parenthesized (message_list_ty *mlp,
 void
 extract_php (FILE *f,
 	     const char *real_filename, const char *logical_filename,
+	     flag_context_list_table_ty *flag_table,
 	     msgdomain_list_ty *mdlp)
 {
   message_list_ty *mlp = mdlp->item[0]->messages;
@@ -1324,6 +1369,8 @@ extract_php (FILE *f,
   last_comment_line = -1;
   last_non_comment_line = -1;
 
+  flag_context_list_table = flag_table;
+
   init_keywords ();
 
   /* Initial mode is HTML mode, not PHP mode.  */
@@ -1331,7 +1378,8 @@ extract_php (FILE *f,
 
   /* Eat tokens until eof is seen.  When extract_parenthesized returns
      due to an unbalanced closing parenthesis, just restart it.  */
-  while (!extract_parenthesized (mlp, -1, 0))
+  while (!extract_parenthesized (mlp, null_context, null_context_list_iterator,
+				 -1, 0))
     ;
 
   /* Close scanner.  */
